/*************************************************************************
|                                                                        |
|   TO.C                                                        07.01.90 |
|                                                                        |
*************************************************************************/

/*************************************************************************
|                                                                        |
|   Written ? by Andrea Steffanoni                                       |
|                                                                        |
|   Modified ? by ? CSM                                                  |
|                                                                        |
|   Modified 7.1.90 by Tim Frost: (BIX: tfrost; CIX: tfrost)             |
|                                                                        |
|       Size of several arrays increased                                 |
|       TODRIVES environment variable added                              |
|       TONORTON environment variable added                              |
|       Run source through CB                                            |
|       Add Watcom specifics                                             |
|       Increase max path length to _MAX_PATH                            |
|       Trap too many subdirectories                                     |
|       Trap malloc failures                                             |
|       Trap too many patterns                                           |
|       Trap too many exact matches                                      |
|       Trap too many partial matches                                    |
|       Allow for single-keystroke selection from more than 9 matches    |
|       Make scantable void, since result was ignored                    |
|       Change all function declarations to new-style                    |
|       Eliminate warnings                                               |
|       Replace existe with access                                       |
|       On-line documentation revised                                    |
|       Some spellings corrected !                                       |
|       DOC file not changed                                             |
|                                                                        |
|   Modified 07.27.94 by William Luitje (luitje@m-net.arbornet.org)      |
|                                                                        |
|       Fixed bug in creating & reading TREEINFO.NCD when there are      |
|        more than 255 directories on a disk.   (Welcome to the world of |
|    big partitions & cheap hard disks!)                                 |
|       Added "approximate" search (-a) using Ratcliff/Obershelp alg.    |
|       Speeded up search by using case sensitive string comparison      |
|        after converting search patterns to upper case (TREEINFO.NCD    |
|        is already uppercase).                                          |
|       Added arbitrarily chosen version number (2.0) to keep track      |
|    of changes                                                          |
|   Added NOFRILLS conditional compilation switch to eliminate           |
|    some "fancier" features. This version uses 1 less block.            |
|    One could also save space by eliminating all calls to printf.       |
|   Added printout of # of directories                                   |
|                                                                        |
*************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <fcntl.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>

#ifdef __TURBOC__

#define COMPILER_SELECTED 1
#include <alloc.h>
#include <dir.h>

#define _dos_getdrive(x)        (*(x) = getdisk() + 1)
#define _dos_setdrive(x, y)     setdisk(x - 1)
#define find_t                  ffblk
#define _A_SUBDIR               FA_DIREC
#define _A_NORMAL               0
#define _dos_findfirst(a, b, c) findfirst(a, c, b)
#define attrib                  ff_attrib
#define name                    ff_name
#define _dos_findnext(x)        findnext(x)
#define _MAX_PATH               MAXPATH

#endif

/* wvl, uncomment this to make a version without push-pop & approx search */
//#define NOFRILLS

#ifdef __WATCOMC__

#define COMPILER_SELECTED 1

#include <malloc.h>
#include <direct.h>

#define strcmpi(x,y)            stricmp(x,y)
#define _dos_getdrive(x)        (*(x) = (0xf & bdos(0x19,0,0)) + 1)
#define _dos_setdrive(x, y)     (*(y) = bdos(0xE,0xf &((x) - 1),0))
#define cdecl

#endif

#ifndef COMPILER_SELECTED 

/* Microsoft C */

#include <malloc.h>
#include <direct.h>
#define  cdecl _cdecl

#endif

char *documentation[] = {
#ifdef NOFRILLS
    "TO - by Andrea Steffanoni et al., version 2.0, No Frills",
#else
    "TO - by Andrea Steffanoni et al., version 2.0, Full",
#endif
    "TO enables you to change directory without entering the full DOS path.",
    "",
    "Syntax:          to [flags] [[*]directory] [[*]directory]....",
    "",
    "Flags (preceded by '-' or '/'):",
    "",
    "    -b           Build (or rebuild) the directory tree file",
    "    -ddrives     Search on the specified drives (eg \"-dCD\")",
    "    -n           Enables use of Norton Change Directory (must be in PATH)",
#ifndef NOFRILLS
    "    -p           Pop to the last directory recorded",
    "    -r           Record actual directory for future popping",
    "    -a           Approximate string matching (don't use with *)",
#endif
    "",
    "`directory' is the name (without path) you want to switch to.  You can supply",
    "a complete name or just its leading characters.  If you precede the name with",
    "`*' then any directory name containing the supplied characters is selected.",
    "",
    "Example: to switch to directory \\lotus\\john\\oldfiles,  enter    to oldfiles",
    "                                                  or simply     to old",
    "                                                    or also     to *files",
#ifndef NOFRILLS
    "                                                    or even     to -a oddfiles",
#endif
    "                        to search on drives C: and D: enter     to -dCD old",
    "",
    "                         --Press any key for more help--",
    0 };

char *patdoc[] = {
    "If TO finds more than one directory matching your specification, it will",
    "display all the matching entries found and wait for your choice.",
    "",
    "TO uses a Directory Tree File, named TREEINFO.NCD, placed in the root",
    "directory.  This file contains details of the directory tree of the disk. If",
    "the file does not exist, TO will create it.  If you have modified your disk",
    "directory structure by adding or deleting directories, use the `-b' flag to",
    "force TO to build a fresh copy of TREEINFO.NCD.",
    "",
    "TREEINFO.NCD is compatible with the Norton Change Directory utility, so if you",
    "have NCD you can continue to use it.  Also, you can use NCD from within TO by",
    "specifying the -n flag in the command line.  NCD will then be called from TO",
    "either if no directory name is given or if TO fails to find any directory",
    "name matching your supplied pattern(s).  NCD.EXE must be on your PATH.",
    "",
    "TO will also look for a file named TO.INI in the root directory of the current",
    "disk, in which you can put one or both of the following command lines:",
    "    DRIVES = [list of drives]          (e.g. \"DRIVES = C D\")",
    "    NORTON = YES",
    "",
    "The first command tells TO to search on all drives specified in the list.  You",
    "can override this command with the -d command-line flag.  The second command is",
    "equivalent to the `-n' flag.  You can also use environment variables TODRIVES",
    "and TONORTON.  For example: \"SET TODRIVES=CD\" or \"SET TONORTON=YES\".",
    0};

#define MAXSUB 128      /* maximum subdirectories processed in each dir */
#define MAXPTN 10       /* maximum number of patterns */
#define MAXEXT 100      /* maximum exact matches */
#define MAXPAR 100      /* maximum partial matches */

struct    t { char nome[13], nomedrive;  int padre; } *table;
int       totdir, exact, partial, ext[MAXEXT], par[MAXPAR];
char      pattern[MAXPTN][15], path[_MAX_PATH], drives[30], drivename[3];
char      actualdir[_MAX_PATH], lastdir[_MAX_PATH];
unsigned  actualdrive;
int       bflag, dflag, nflag, pflag, rflag, aflag;
int       gotpattern = 0;

/*************************************************************************
|                                                                        |
|   Prototypes                                                           |
|                                                                        |
*************************************************************************/

void cdecl     main(int, char *[]);
void           switches(int, char *[]);
void           maketable(void);
unsigned int   scandir(FILE *, int);
void           readtable(void);
void           scantable(int);
int            selectdir(void);
void           readparms(void);
char *         faipath(int);
void           fadrivename(int);
void           help(char * *);
void           usage(char *);
char *         legge(FILE *, char *);
void           scrive(FILE *, char *);
void           norton(void);
void           goodexit(int);
void           badexit(void);
void           salvadir(void);
void           salvaunadir(void);
int            simil(char *, char *);

/*************************************************************************
|                                                                        |
|   main                                                                 |
|                                                                        |
*************************************************************************/

void cdecl main(int argc, char * argv[])

{
    register char * p;
    register int  i;
    int c;

    strcpy(lastdir, "");
    bflag = dflag = nflag = pflag = rflag = aflag = 0;
    _dos_getdrive(&actualdrive);
    getcwd(actualdir,_MAX_PATH);

    switches(argc,argv);
    readparms();

    p = getenv("TODRIVES");
    if (p != NULL)
    {
        if(!dflag)
        {
            dflag = strlen(p);
            strcpy(drives,p);
        }
    }
    p = getenv("TONORTON");
    if (p != NULL)
    {
        if(!strcmpi(p,"YES"))
        {
            nflag = 1;
        }
    }

#ifndef NOFRILLS
    if (pflag)
    {
        if (!lastdir[0])
        {
            printf("\nTO - No directory recorded.");
            badexit();
        }
        strcpy(path, lastdir);
        table = NULL;
        goodexit(1);
    }
#endif

    if (!dflag)
    {
        dflag++;
        _dos_getdrive(&c);
        drives[0] = (char)c;
    }
    else
    {
        strupr(drives);
        for (i = 0; i < dflag; i++)
            drives[i] = (char)(drives[i] - 'A' + 1);
    }

#ifndef NOFRILLS
    if (rflag) salvadir();
#endif

    if (!gotpattern)
    {
        if (nflag) norton();
        usage("No arguments.");
    }

    for (i = 0; i < dflag; i++)
    {
        _dos_setdrive(drives[i], &c);
        fadrivename(i);
        if (access("\\TREEINFO.NCD",0) || bflag)
            maketable();
    }

    for (i = 0; i < dflag; i++)
    {
        _dos_setdrive(drives[i], &c);
        fadrivename(i);
        readtable();
    }
    for (i = 0; i < gotpattern; i++) scantable(i);

    if (exact || partial)
    {
        faipath(selectdir());
        goodexit(1);
    }

    printf("\nTO - Directory not found.\n");

    if (nflag)
    {
        printf("\n     Press N to use Norton Change Directory or ESC to abort.\n");
        while (1)
        {
            if ((c = getch()) == 27)
                badexit();
            if (c == 'n' || c == 'N')
                norton();
            printf("\a");
        }
    }
    badexit();
}

/*************************************************************************
|                                                                        |
|   switches                                                             |
|                                                                        |
*************************************************************************/

void switches(int argc, char * argv[])

{
    register char * p;
    register int  i;
    int c;

    for (i = 1; i < argc; ++i)
    {
        p = argv[i];
        if (*p == '?')
        {
            help(documentation);
            getch();
            help(patdoc);
            exit(0);
        }
        if (*p == '-' || *p == '/')
        {
            ++p;
            while ((c = *p++) != '\0')
            {
                switch(tolower(c))
                {
                case 'b':
                    ++bflag;
                    break;
                case 'd':
                    if (!p)
                        usage("Drive(s) specification missing.");
                    strcpy(drives, p);
                    dflag = strlen(drives);
                    *p = 0;
                    break;
                case 'n':
                    ++nflag;
                    break;
#ifndef NOFRILLS
        case 'p':
                    ++pflag;
                    break;
                case 'r':
                    ++rflag;
                    break;
                case 'a':
                    ++aflag;
                    break;
#endif
        default:
                    usage("Unknown flag.");
                }
            }
        }
        else
        {
            if (strstr(argv[i], "\\") && strlen(argv[i]) > 1)
                usage("Do not use backslash in the directory name.");
            if (gotpattern < MAXPTN) strcpy(pattern[gotpattern++], strupr(argv[i]));
        }
    }
}

/*************************************************************************
|                                                                        |
|   maketable()                                                          |
|                                                                        |
*************************************************************************/

void maketable(void)

{
    int i;
    unsigned int totale = 0;
    unsigned char tot1, tot2;
    char buffer[16];
    FILE * f;

    if ((f = fopen("\\TREEINFO.NCD", "wb")) == NULL)
    {
        printf("TO - Open failure for file %s\\TREEINFO.NCD.\n", drivename);
        badexit();
    }

    printf("\nTO - Creating Directory Tree File %s\\TREEINFO.NCD...", drivename);

    chdir("\\");
    totdir = 1;

    strcpy(buffer, "PNCI");
    buffer[5] = 0;
    buffer[6] = 0;
    buffer[7] = 0;
    buffer[8] = 1;
    fwrite(buffer, 1, 9, f);

    memset(buffer, 0, 16);
    buffer[0]  = '\\';
    buffer[14] = 1;
    fwrite(buffer, 1, 16, f);
    for (i = 0; i < 16; i++)
        totale += (unsigned int)buffer[i];

    totale += scandir(f, 1);


    tot1 = (unsigned char)(totale / 256);
    tot2 = (unsigned char)(totale - (totale / 256) * 256);
    buffer[0] = tot2;
    buffer[1] = tot1;
    fwrite(buffer, 1, 2, f);
    fseek(f, 5L, SEEK_SET);
    buffer[0] = (char)(totdir & 0xff);
    buffer[1] = (char)(totdir / 256);
    buffer[2] = (char)(totdir + 42);
    fwrite(buffer, 1, 3, f);
    fclose(f);
    printf("\b\b\b with %d entries.\n",totdir);
    totdir = 0;
}

/*************************************************************************
|                                                                        |
|   scandir(f, liv)                                                      |
|                                                                        |
*************************************************************************/

unsigned int scandir(FILE * f, int liv)

{
    register i, i1;
    char buffer[20], * dir[MAXSUB];
#ifdef __WATCOMC__
    DIR * dirp;
    struct dirent * direntp;
#else
    struct find_t files_in_dir;
#endif
    int num = 0;
    int prima = 1;
    unsigned int totale = 0;

    memset(buffer, 0, 16);

#ifdef __WATCOMC__
    dirp = opendir("*.*");
    if (dirp != NULL)
    {
        while (1)
        {
            direntp = readdir(dirp);
            if (direntp == NULL) break;
            if ((direntp->d_attr & _A_SUBDIR) && (direntp->d_name[0] != '.'))
            {
                if ((dir[num] = malloc(13)) == NULL)
                {
                    printf("TO - Out of memory.\n");
                    badexit();
                }
                strcpy(dir[num++], direntp->d_name);
            }
        }
        closedir(dirp);
    }
#else
    if (!_dos_findfirst("*.*", _A_SUBDIR, &files_in_dir))
    {
        if (files_in_dir.attrib == _A_SUBDIR && files_in_dir.name[0] != '.')
        {
            if ((dir[num] = malloc(13)) == NULL)
            {
                printf("TO - Out of memory.\n");
                badexit();
            }
            strcpy(dir[num++], files_in_dir.name);
        }
        while (!_dos_findnext(&files_in_dir))
        {
            if (files_in_dir.attrib == _A_SUBDIR && files_in_dir.name[0] != '.')
            {
                if ((dir[num] = malloc(13)) == NULL)
                {
                    printf("TO - Out of memory.\n");
                    badexit();
                }
                strcpy(dir[num++], files_in_dir.name);
            }
            if (num >= MAXSUB)
            {
                printf("TO - Too many subdirectories.\n");
                badexit();
            }
        }
    }
#endif

    for (i = 0; i < num; i++)
    {
        memset(buffer, 0, 16);
        strcpy(buffer, dir[i]);
        buffer[13] = (char)liv;
        buffer[14] = (char)prima;
        if (i != num - 1)
            buffer[15] = 1;
        fwrite(buffer, 1, 16, f);
        for (i1 = 0; i1 < 16; i1++)
            totale += (unsigned int)buffer[i1];
        totdir++;
        prima = 0;
        chdir(dir[i]);
        free(dir[i]);
        totale += scandir(f, liv + 1);
    }

    chdir("..");
    return(totale);
}

/*************************************************************************
|                                                                        |
|   readtable()                                                          |
|                                                                        |
*************************************************************************/

void readtable(void)

{
    register i;
    int f, totqui, inizio, livello, padri[20];
    unsigned char buffer[16];       /* CMS Changed This */

    if ((f = open("\\TREEINFO.NCD", (int)(O_RDONLY | O_BINARY),0)) == -1)
    {
        printf("TO - Open failure for file %s\\TREEINFO.NCD.\n", drivename);
        badexit();
    }
    lseek(f, 5L, SEEK_SET);
    read(f, buffer, 2);
    totqui = (int)buffer[0] + 256 * buffer[1];
    inizio = totdir + 1;
    if (!totdir)
        table = (struct t *)malloc((totqui + 1) * sizeof(struct t));
    else
        table = (struct t *)realloc(table, (totdir + totqui + 1) * sizeof(struct t));
    if (table == NULL)
    {
        printf("TO - Out of memory.\n");
        badexit();
    }
    totdir += totqui;

    lseek(f, 9L, SEEK_SET);
    for (i = inizio; i <= totdir; i++)
    {
        read(f, buffer, 16);
        strcpy(table[i].nome, buffer);
        table[i].nomedrive = drivename[0];
        livello = (int)buffer[13];
        padri[livello] = i;
        if (livello)
            table[i].padre = padri[livello - 1];
        else
            table[i].padre = 0;
    }
    close(f);
}

/*************************************************************************
|                                                                        |
|   scantable(num)                                                       |
|                                                                        |
*************************************************************************/

void scantable(int num)

{
    register int i;
    int j, k, percent, kpart, inserted;
    #define MAXCAN 9
    char candidates[MAXCAN][13];
    int scoreCan[MAXCAN], indexCan[MAXCAN];

    for (i = 1; i <= totdir; i++)
    {
        if (pattern[num][0] == '*')
        {
            if (strstr(table[i].nome, &(pattern[num][1])))
            {
                if (partial < MAXPAR) par[partial++] = i;
            }
            continue;
        }
#ifndef NOFRILLS
/* begin approximate match section, wvl Sep 92 */
        if(aflag) {
            percent = simil(pattern[num],table[i].nome);
            if(percent == 100) {                    /* check for exact match */
                if(exact < MAXEXT) ext[exact++] = i;
                continue; }
            /* only worry about partial matches if no exact matches and some similarity */
            if(exact == 0 && percent !=0) {
                inserted = 0;
                kpart = partial; /*??*/
                for( j=0; j<partial; ++j) {
                    if(percent > scoreCan[j]) {
                        for( k=MAXCAN-1; k>j; --k) { /* move lower ranked candidates down */
                            strcpy(candidates[k],candidates[k-1]);
                            scoreCan[k] = scoreCan[k-1];
                            indexCan[k] = indexCan[k-1]; }
                        strcpy(candidates[j],table[i].nome); /* insert new candidate here */
                        scoreCan[j] = percent;
                        indexCan[j] = i;
                        inserted = 1;
                        if(partial < MAXCAN)
                            partial++;
                        break;
                    }
                }
                if(!inserted && partial < MAXCAN) {         /* if not inserted & room, add to end of table */
                    strcpy(candidates[partial],table[i].nome);
                    scoreCan[partial] = percent;
                    indexCan[partial++] = i; }
                    continue;
            }
        }
/* end approximate match section, wvl */
#endif
    if (!strcmp(pattern[num], table[i].nome))
        {
            if (exact < MAXEXT) ext[exact++] = i;
            continue;
        }
        if (!strncmp(pattern[num], table[i].nome, strlen(pattern[num])))
        {
            if (partial < MAXPAR) par[partial++] = i;
        }
    }

    if(!exact && aflag) /* move similarity candidate indices to partial match array in order of decreasing score */
        for( j=0; j<partial; ++j)
            par[j] = indexCan[j];
}


/*************************************************************************
|                                                                        |
|   selectdir()                                                          |
|                                                                        |
*************************************************************************/

int selectdir(void)

{
    register i;
    int l;
    int start, adjust;

    if (exact == 1) return(ext[0]);
    if (!exact && partial == 1) return(par[0]);
    printf("\n");

    if (exact)
    {
        start = 0;
        while (1)
        {
            adjust = start;
            printf("  Please choose one or ESC to abort:\n\n");
            for (i = 0 ; (i + start) < exact; i++)
            {
                printf("    %d - %s\n", i + 1, faipath(ext[i + adjust]));
                if ((i >= 8) && ((start + 9) < exact))
                {
                    printf("\n  Key <sp> for more\n\n");
                    start += 9;
                    break;
                }
            }
            while (1)
            {
                if ((l = getch()) == 27)
                    badexit();
                else if (start && (l == 32))
                    break;
                else
                {
                    if ((l >= '1') && (l <= '9')) l -= '1';
                    if (l < i) return ext[l + adjust];
                    printf("\a");
                }
            }
        }
    }

    start = 0;
    while (1)
    {
        adjust = start;
        printf("  Please choose one or ESC to abort:\n\n");
        for (i = 0; (i + start) < partial; i++)
        {
            printf("    %d - %s\n", i + 1, faipath(par[i + adjust]));
            if ((i >= 8) && ((start + 9) < partial))
            {
                printf("\n  Key <sp> for more\n\n");
                adjust = start;
                start += 9;
                break;
            }
        }
        while (1)
        {
            if ((l = getch()) == 27)
                badexit();
            else if (start && (l == 32))
                break;
            else
            {
                if ((l >= '1') && (l <= '9')) l -= '1';
                if (l < i) return par[l + adjust];
                printf("\a");
            }
        }
    }
}

/*************************************************************************
|                                                                        |
|   faipath(num)                                                         |
|                                                                        |
*************************************************************************/

char * faipath(int num)

{
    int tot = 0;
    int i, chain[20];

    while (table[num].padre)
    {
        chain[tot++] = num;
        num = table[num].padre;
    }

    path[0] = table[num].nomedrive;
    path[1] = 0;
    strcat(path, ":\\");
    for (i = tot; i; i--)
    {
        strcat(path, strlwr(table[chain[i - 1]].nome));
        if (i > 1)
            strcat(path, "\\");
    }
    return(path);
}

/*************************************************************************
|                                                                        |
|   readparms()                                                          |
|                                                                        |
*************************************************************************/

void readparms(void)

{
    register i;
    int l;
    FILE *f;
    char buffer[_MAX_PATH], command[_MAX_PATH], *p;

    if (access("\\TO.INI",0)) return;

    if ((f = fopen("\\TO.INI", "rt")) == NULL)
    {
        printf("TO - Open failure for file \\TO.INI.\n");
        badexit();
    }

    while (!feof(f))
    {
        legge(f, buffer);
        if ((p = strchr(buffer, ';')) != NULL)
            *p = '\0';
        for (i = 0, l = 0; i < strlen(buffer); i++)
        {
            if (buffer[i] != ' ')
                command[l++] = buffer[i];
        }
        command[l] = 0;
        if (!command[0])
            continue;
        if (!strnicmp(command, "drives=", 7))
        {
            if (!dflag)
            {   dflag = strlen(command) - 7;
                strcpy(drives, command + 7);
            }
            continue;
        }
        if (!strnicmp(command, "norton=yes", 10))
        {
            nflag++;
            continue;
        }
        if (!strnicmp(command, "lastdir=", 8))
        {
            strcpy(lastdir, command + 8);
            continue;
        }
        printf("\nTO - Unknown command in \\TO.INI; ignored.\n");
    }
    fclose(f);
}

/*************************************************************************
|                                                                        |
|   fadrivename(num)                                                     |
|                                                                        |
*************************************************************************/

void fadrivename(int num)

{
    drivename[0] = (char)(drives[num] + 'A' - 1);
    drivename[1] = ':';
    drivename[2] = '\0';
}

/*************************************************************************
|                                                                        |
|   help(hp)                                                             |
|                                                                        |
*************************************************************************/

void help(char * * hp)

{
    register char * * dp;

    for (dp = hp; *dp; dp++)
        printf("\n%s", *dp);
}

/*************************************************************************
|                                                                        |
|   usage(s)                                                             |
|                                                                        |
*************************************************************************/

void usage(char *s)

{
    fprintf(stderr, "\nTO 2.0 - %s\n\n", s);
#ifdef NOFRILLS
    fprintf(stderr, "Usage:     to [-b] [-ddriveletters] [-n] [[*]directory ...]\n");
#else
    fprintf(stderr, "Usage:     to [-b] [-ddriveletters] [-n] [-p] [-r] [-a] [[*]directory ...]\n");
#endif
    fprintf(stderr, "For help:  to ?\n");
    badexit();
}

/*************************************************************************
|                                                                        |
|   legge(archivio, stringa)                                             |
|                                                                        |
|   Legge una stringa (gia' allocata!) dall'archivio (modo testo!) e     |
|   toglie il \r\n finale.  Ritorna il puntatore alla stringa letta.     |
|                                                                        |
*************************************************************************/

char *legge(FILE * archivio, char * stringa)

{
    fgets(stringa, _MAX_PATH, archivio);
    if (strlen(stringa))
        stringa[strlen(stringa) - 1] = '\0';
    if (strlen(stringa) && stringa[strlen(stringa) - 1] == '\r')
        stringa[strlen(stringa) - 1] = '\0';
    return(stringa);
}

/*************************************************************************
|                                                                        |
|   scrive(archivio, stringa)                                            |
|                                                                        |
*************************************************************************/

void scrive(FILE * archivio, char * stringa)

{
    strcat(stringa, "\r\n");
    fputs(stringa, archivio);
}

/*************************************************************************
|                                                                        |
|   norton()                                                             |
|                                                                        |
*************************************************************************/

void norton(void)

{
    register int i;
    int c;
    char dirqui[_MAX_PATH], dirnow[_MAX_PATH];

    for (i = 0; i < dflag; i++)
    {
        _dos_setdrive(drives[i], &c);
        getcwd(dirqui,_MAX_PATH);
        system("NCD");
        getcwd(dirnow,_MAX_PATH);
        if (strcmp(dirqui, dirnow))
            goodexit(0);
    }
    badexit();
}

/*************************************************************************
|                                                                        |
|   goodexit(modo)                                                       |
|                                                                        |
*************************************************************************/

void goodexit(int modo)

{
    int c;

    if (modo)
    {
        _dos_setdrive((unsigned)(path[0] - 'A' + 1), &c);
        chdir(path + 2);
    }
    exit(0);
}

/*************************************************************************
|                                                                        |
|   badexit()                                                            |
|                                                                        |
*************************************************************************/

void badexit(void)

{
    int c;

    _dos_setdrive(actualdrive, &c);
    chdir(actualdir);
    exit(1);
}

#ifndef NOFRILLS
/*************************************************************************
|                                                                        |
|   salvadir()                                                           |
|                                                                        |
*************************************************************************/

void salvadir(void)

{
    int i, c;

    for (i = 0, c = 10; i < dflag; i++)
    {
        _dos_setdrive(drives[i], &c);
        salvaunadir();
    }
    _dos_setdrive(actualdrive, &c);
}

/*************************************************************************
|                                                                        |
|   salvaunadir()                                                        |
|                                                                        |
*************************************************************************/

void salvaunadir(void)

{
    register i;
    int l;
    FILE *f, *f2;
    char buffer[_MAX_PATH], command[_MAX_PATH];

    if (!access("\\TO.INI",0))
    {
        if ((f = fopen("\\TO.INI", "rt")) == NULL)
        {
            printf("TO - Open failure for file \\TO.INI.\n");
            badexit();
        }
        if ((f2 = fopen("\\TO.TMP", "wt")) == NULL)
        {
            printf("TO - Open failure for file \\TO.TMP.\n");
            badexit();
        }

        while (1)
        {
            legge(f, buffer);
            if (feof(f))
                break;
            for (i = 0, l = 0; i < strlen(buffer); i++)
            {
                if (buffer[i] != ' ')
                    command[l++] = buffer[i];
            }
            command[l] = 0;
            if (strnicmp(command, "lastdir=", 8))
                scrive(f2, buffer);
        }

        fclose(f);
        fclose(f2);
    }

    remove("\\TO.INI");
    rename("\\TO.TMP", "\\TO.INI");

    if ((f2 = fopen("\\TO.INI", "at")) == NULL)
    {
        printf("TO - Open failure for file \\TO.INI.\n");
        badexit();
    }
    strcpy(buffer, "LASTDIR = ");
    strcat(buffer, actualdir);
    strcat(buffer, "  ;last directory recorded");
    scrive(f2, buffer);
    fclose(f2);
}

int GCsubstr(char *, char *, char *, char *);
/* Ratcliff/Obershelp pattern matching
 *
 *  Approximate word matching: takes two words and returns a
 *  similarity score based on co-occurrence of subpatterns.
 *
 *  This code appeared in a letter to the editor in DDJ, 11/88.
 *  The original article on the pattern matching, "Pattern Matching
 *  by Gestalt" by John Ratcliff, appeared in the July 1988
 *  issue (#181) but the algorithm was presented in assembly.  
 *
 *  The 11/88 issue also contained another verison in C which was
 *  a more faithful translation of the original; it has the
 *  advantage of not being recursive.
 *
 *  The algorithm seems to work nicely for a variety of test cases.
 *  The main drawback of the algorithm is the cost of the pairwise
 *  comparisons.  It is significantly more expensive than stemming,
 *  soundex, and the like.  Might consider using this as a second
 *  phase...
 */

int simil(s1, s2)
 char *s1, *s2;
{
 short l1, l2;

    l1 = strlen(s1);
    l2 = strlen(s2);
    
    /* exact match end-case */
    if( l1 == 1 && l2 == 1 && *s1 == *s2 )  return(100);
            
    return(200 * GCsubstr(s1, s1 + l1, s2, s2 + l2) / (l1 + l2));
}

int GCsubstr(st1, end1, st2, end2)
 char *st1, *end1, *st2, *end2;
{
 register char *a1, *a2;
 char *b1, *s1, *b2, *s2;
 short max, i;

    if( end1 <= st1 || end2 <= st2 ) return(0);
    if( end1 == st1 + 1 && end2 == st2 + 1 ) return(0);
        
    max = 0;
    b1 = end1; b2 = end2;
    
    for( a1 = st1; a1 < b1; a1++ ) {
        for( a2 = st2; a2 < b2; a2++ ) {
            if( *a1 == *a2 ) {
                /* determine length of common substring */
                for( i = 1; a1[i] && (a1[i] == a2[i]); i++ ) 
                    ;
                if( i > max ) {
                    max = i; s1 = a1; s2 = a2;
                    b1 = end1 - max; b2 = end2 - max;
                }
            }
        }
    }
    if( !max ) return(0);
    max += GCsubstr(s1 + max, end1, s2 + max, end2);    /* rhs */
    max += GCsubstr(st1, s1, st2, s2);          /* lhs */
    return(max);
}
#endif
